"""Module for XKNX Exception handling."""
# flake8: noqa
from .exception import (
    ConversionError,
    CouldNotParseAddress,
    CouldNotParseKNXIP,
    CouldNotParseTelegram,
    DeviceIllegalValue,
    UnsupportedCEMIMessage,
    XKNXException,
)
